<?php
/**
 * Copyright (c) 2012-2016, Agence Française Informatique (AFI). All rights reserved.
 *
 * LINSE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE as published by
 * the Free Software Foundation.
 *
 * There are special exceptions to the terms and conditions of the AGPL as it
 * is applied to this software (see README file).
 *
 * LINSE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU AFFERO GENERAL PUBLIC LICENSE for more details.
 *
 * You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 * along with LINSE; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
 */


class Linse_Iso5426Converter extends Linse_Converter {
  public function toIndexable($data, $compact=true) {
    $data = preg_replace('/\210.+\211/', '', $data); // Les delimiteurs d'article bnf NSB NSE

    $result = '';
    $len = strlen($data);
    for ($i=0; $i < $len; $i++) {
      $ord = ord($data[$i]);
      if (0xe0 < $ord) {
        $result .= $this->charToIndexable($ord);
        continue;
      }

      $result .= $ord == 0x80 ? '€' : $data[$i];
    }

    $result = preg_replace('/[' . chr(123) . '-' . chr(255) .']/', '', $result);
    return parent::toIndexable($result, $compact);
  }


  public function charToIndexable($c) {
    $map = [0xe1 => 'AE',
            0xe2 => 'D',
            0xe6 => 'IJ',
            0xe8 => 'L',
            0xe9 => 'O',
            0xea => 'OE',
            0xec => 'TH',
            0xf1 => 'ae',
            0xf2 => 'd',
            0xf3 => 'd',
            0xf5 => 'i',
            0xf6 => 'ij',
            0xf8 => 'l',
            0xf9 => 'o',
            0xfa => 'oe',
            0xfb => 'ss',
            0xfc => 'th'];

    return array_key_exists($c, $map) ? $map[$c] : $c;
  }


  public function toUnicode($data) {
    $data = str_replace([chr(136), chr(137)], '', $data); // Les delimiteurs d'article bnf

    if (!preg_match("/[\xC0-\xFF]/misU", $data))
      return $data;

    $result = '';
    $len = strlen($data);
    for ($i=0; $i < $len; $i++) {
      if (ord($data[$i]) >= 0xC1) {
        $result .= $this->charsToUnicode(ord($data[$i]), ord($data[$i+1]));
        $i++;
        continue;
      }

      $result .= (ord($data[$i]) == 0x80) ? '€' : $data[$i];
    }

    return $result;
  }


  public function charsToUnicode($c1, $c2) {
    $combined = [];

    $combined[0xc041] = 0x1ea2;    # capital a with hook above
    $combined[0xc045] = 0x1eba;    # capital e with hook above
    $combined[0xc049] = 0x1ec8;    # capital i with hook above
    $combined[0xc04f] = 0x1ece;    # capital o with hook above
    $combined[0xc055] = 0x1ee6;    # capital u with hook above
    $combined[0xc059] = 0x1ef6;    # capital y with hook above
    $combined[0xc061] = 0x1ea3;    # small a with hook above
    $combined[0xc065] = 0x1ebb;    # small e with hook above
    $combined[0xc069] = 0x1ec9;    # small i with hook above
    $combined[0xc06f] = 0x1ecf;    # small o with hook above
    $combined[0xc075] = 0x1ee7;    # small u with hook above
    $combined[0xc079] = 0x1ef7;    # small y with hook above

    # 4/1 grave accent
    $combined[0xc141] = 0x00c0;    # capital a with grave accent
    $combined[0xc145] = 0x00c8;    # capital e with grave accent
    $combined[0xc149] = 0x00cc;    # capital i with grave accent
    $combined[0xc14f] = 0x00d2;    # capital o with grave accent
    $combined[0xc155] = 0x00d9;    # capital u with grave accent
    $combined[0xc157] = 0x1e80;    # capital w with grave
    $combined[0xc159] = 0x1ef2;    # capital y with grave
    $combined[0xc161] = 0x00e0;    # small a with grave accent
    $combined[0xc165] = 0x00e8;    # small e with grave accent
    $combined[0xc169] = 0x00ec;    # small i with grave accent
    $combined[0xc16f] = 0x00f2;    # small o with grave accent
    $combined[0xc175] = 0x00f9;    # small u with grave accent
    $combined[0xc177] = 0x1e81;    # small w with grave
    $combined[0xc179] = 0x1ef3;    # small y with grave
    # 4/2 acute accent
    $combined[0xc241] = 0x00c1;    # capital a with acute accent
    $combined[0xc243] = 0x0106;    # capital c with acute accent
    $combined[0xc245] = 0x00c9;    # capital e with acute accent
    $combined[0xc247] = 0x01f4;    # capital g with acute
    $combined[0xc249] = 0x00cd;    # capital i with acute accent
    $combined[0xc24b] = 0x1e30;    # capital k with acute
    $combined[0xc24c] = 0x0139;    # capital l with acute accent
    $combined[0xc24d] = 0x1e3e;    # capital m with acute
    $combined[0xc24e] = 0x0143;    # capital n with acute accent
    $combined[0xc24f] = 0x00d3;    # capital o with acute accent
    $combined[0xc250] = 0x1e54;    # capital p with acute
    $combined[0xc252] = 0x0154;    # capital r with acute accent
    $combined[0xc253] = 0x015a;    # capital s with acute accent
    $combined[0xc255] = 0x00da;    # capital u with acute accent
    $combined[0xc257] = 0x1e82;    # capital w with acute
    $combined[0xc259] = 0x00dd;    # capital y with acute accent
    $combined[0xc25a] = 0x0179;    # capital z with acute accent
    $combined[0xc261] = 0x00e1;    # small a with acute accent
    $combined[0xc263] = 0x0107;    # small c with acute accent
    $combined[0xc265] = 0x00e9;    # small e with acute accent
    $combined[0xc267] = 0x01f5;    # small g with acute
    $combined[0xc269] = 0x00ed;    # small i with acute accent
    $combined[0xc26b] = 0x1e31;    # small k with acute
    $combined[0xc26c] = 0x013a;    # small l with acute accent
    $combined[0xc26d] = 0x1e3f;    # small m with acute
    $combined[0xc26e] = 0x0144;    # small n with acute accent
    $combined[0xc26f] = 0x00f3;    # small o with acute accent
    $combined[0xc270] = 0x1e55;    # small p with acute
    $combined[0xc272] = 0x0155;    # small r with acute accent
    $combined[0xc273] = 0x015b;    # small s with acute accent
    $combined[0xc275] = 0x00fa;    # small u with acute accent
    $combined[0xc277] = 0x1e83;    # small w with acute
    $combined[0xc279] = 0x00fd;    # small y with acute accent
    $combined[0xc27a] = 0x017a;    # small z with acute accent
    $combined[0xc2e1] = 0x01fc;    # capital ae with acute
    $combined[0xc2f1] = 0x01fd;    # small ae with acute
    # 4/3 circumflex accent
    $combined[0xc341] = 0x00c2;    # capital a with circumflex accent
    $combined[0xc343] = 0x0108;    # capital c with circumflex
    $combined[0xc345] = 0x00ca;    # capital e with circumflex accent
    $combined[0xc347] = 0x011c;    # capital g with circumflex
    $combined[0xc348] = 0x0124;    # capital h with circumflex
    $combined[0xc349] = 0x00ce;    # capital i with circumflex accent
    $combined[0xc34a] = 0x0134;    # capital j with circumflex
    $combined[0xc34f] = 0x00d4;    # capital o with circumflex accent
    $combined[0xc353] = 0x015c;    # capital s with circumflex
    $combined[0xc355] = 0x00db;    # capital u with circumflex
    $combined[0xc357] = 0x0174;    # capital w with circumflex
    $combined[0xc359] = 0x0176;    # capital y with circumflex
    $combined[0xc35a] = 0x1e90;    # capital z with circumflex
    $combined[0xc361] = 0x00e2;    # small a with circumflex accent
    $combined[0xc363] = 0x0109;    # small c with circumflex
    $combined[0xc365] = 0x00ea;    # small e with circumflex accent
    $combined[0xc367] = 0x011d;    # small g with circumflex
    $combined[0xc368] = 0x0125;    # small h with circumflex
    $combined[0xc369] = 0x00ee;    # small i with circumflex accent
    $combined[0xc36a] = 0x0135;    # small j with circumflex
    $combined[0xc36e] = 0x00f1;    # small n with tilde
    $combined[0xc36f] = 0x00f4;    # small o with circumflex accent
    $combined[0xc373] = 0x015d;    # small s with circumflex
    $combined[0xc375] = 0x00fb;    # small u with circumflex
    $combined[0xc377] = 0x0175;    # small w with circumflex
    $combined[0xc379] = 0x0177;    # small y with circumflex
    $combined[0xc37a] = 0x1e91;    # small z with circumflex
    # 4/4 tilde
    $combined[0xc441] = 0x00c3;    # capital a with tilde
    $combined[0xc445] = 0x1ebc;    # capital e with tilde
    $combined[0xc449] = 0x0128;    # capital i with tilde
    $combined[0xc44e] = 0x00d1;    # capital n with tilde
    $combined[0xc44f] = 0x00d5;    # capital o with tilde
    $combined[0xc455] = 0x0168;    # capital u with tilde
    $combined[0xc456] = 0x1e7c;    # capital v with tilde
    $combined[0xc459] = 0x1ef8;    # capital y with tilde
    $combined[0xc461] = 0x00e3;    # small a with tilde
    $combined[0xc465] = 0x1ebd;    # small e with tilde
    $combined[0xc469] = 0x0129;    # small i with tilde
    $combined[0xc46e] = 0x00f1;    # small n with tilde
    $combined[0xc46f] = 0x00f5;    # small o with tilde
    $combined[0xc475] = 0x0169;    # small u with tilde
    $combined[0xc476] = 0x1e7d;    # small v with tilde
    $combined[0xc479] = 0x1ef9;    # small y with tilde
    # 4/5 macron
    $combined[0xc541] = 0x0100;    # capital a with macron
    $combined[0xc545] = 0x0112;    # capital e with macron
    $combined[0xc547] = 0x1e20;    # capital g with macron
    $combined[0xc549] = 0x012a;    # capital i with macron
    $combined[0xc54f] = 0x014c;    # capital o with macron
    $combined[0xc555] = 0x016a;    # capital u with macron
    $combined[0xc561] = 0x0101;    # small a with macron
    $combined[0xc565] = 0x0113;    # small e with macron
    $combined[0xc567] = 0x1e21;    # small g with macron
    $combined[0xc569] = 0x012b;    # small i with macron
    $combined[0xc56f] = 0x014d;    # small o with macron
    $combined[0xc575] = 0x016b;    # small u with macron
    $combined[0xc572] = 0x0159;    # small r with macron
    $combined[0xc5e1] = 0x01e2;    # capital ae with macron
    $combined[0xc5f1] = 0x01e3;    # small ae with macron
    # 4/6 breve
    $combined[0xc641] = 0x0102;    # capital a with breve
    $combined[0xc645] = 0x0114;    # capital e with breve
    $combined[0xc647] = 0x011e;    # capital g with breve
    $combined[0xc649] = 0x012c;    # capital i with breve
    $combined[0xc64f] = 0x014e;    # capital o with breve
    $combined[0xc655] = 0x016c;    # capital u with breve
    $combined[0xc661] = 0x0103;    # small a with breve
    $combined[0xc665] = 0x0115;    # small e with breve
    $combined[0xc667] = 0x011f;    # small g with breve
    $combined[0xc669] = 0x012d;    # small i with breve
    $combined[0xc66f] = 0x014f;    # small o with breve
    $combined[0xc675] = 0x016d;    # small u with breve
    # 4/7 dot above
    $combined[0xc7b0] = 0x01e1;    # Ain with dot above
    $combined[0xc742] = 0x1e02;    # capital b with dot above
    $combined[0xc743] = 0x010a;    # capital c with dot above
    $combined[0xc744] = 0x1e0a;    # capital d with dot above
    $combined[0xc745] = 0x0116;    # capital e with dot above
    $combined[0xc746] = 0x1e1e;    # capital f with dot above
    $combined[0xc747] = 0x0120;    # capital g with dot above
    $combined[0xc748] = 0x1e22;    # capital h with dot above
    $combined[0xc749] = 0x0130;    # capital i with dot above
    $combined[0xc74d] = 0x1e40;    # capital m with dot above
    $combined[0xc74e] = 0x1e44;    # capital n with dot above
    $combined[0xc750] = 0x1e56;    # capital p with dot above
    $combined[0xc752] = 0x1e58;    # capital r with dot above
    $combined[0xc753] = 0x1e60;    # capital s with dot above
    $combined[0xc754] = 0x1e6a;    # capital t with dot above
    $combined[0xc757] = 0x1e86;    # capital w with dot above
    $combined[0xc758] = 0x1e8a;    # capital x with dot above
    $combined[0xc759] = 0x1e8e;    # capital y with dot above
    $combined[0xc75a] = 0x017b;    # capital z with dot above
    $combined[0xc761] = 0x0227;    # small b with dot above
    $combined[0xc762] = 0x1e03;    # small b with dot above
    $combined[0xc763] = 0x010b;    # small c with dot above
    $combined[0xc764] = 0x1e0b;    # small d with dot above
    $combined[0xc765] = 0x0117;    # small e with dot above
    $combined[0xc766] = 0x1e1f;    # small f with dot above
    $combined[0xc767] = 0x0121;    # small g with dot above
    $combined[0xc768] = 0x1e23;    # small h with dot above
    $combined[0xc76d] = 0x1e41;    # small m with dot above
    $combined[0xc76e] = 0x1e45;    # small n with dot above
    $combined[0xc770] = 0x1e57;    # small p with dot above
    $combined[0xc772] = 0x1e59;    # small r with dot above
    $combined[0xc773] = 0x1e61;    # small s with dot above
    $combined[0xc774] = 0x1e6b;    # small t with dot above
    $combined[0xc777] = 0x1e87;    # small w with dot above
    $combined[0xc778] = 0x1e8b;    # small x with dot above
    $combined[0xc779] = 0x1e8f;    # small y with dot above
    $combined[0xc77a] = 0x017c;    # small z with dot above
    # 4/8 trema, diaresis
    $combined[0xc820] = 0x00a8;    # diaeresis
    $combined[0xc841] = 0x00c4;    # capital a with diaeresis
    $combined[0xc845] = 0x00cb;    # capital e with diaeresis
    $combined[0xc848] = 0x1e26;    # capital h with diaeresis
    $combined[0xc849] = 0x00cf;    # capital i with diaeresis
    $combined[0xc84f] = 0x00d6;    # capital o with diaeresis
    $combined[0xc855] = 0x00dc;    # capital u with diaeresis
    $combined[0xc857] = 0x1e84;    # capital w with diaeresis
    $combined[0xc858] = 0x1e8c;    # capital x with diaeresis
    $combined[0xc859] = 0x0178;    # capital y with diaeresis
    $combined[0xc861] = 0x00e4;    # small a with diaeresis
    $combined[0xc865] = 0x00eb;    # small e with diaeresis
    $combined[0xc868] = 0x1e27;    # small h with diaeresis
    $combined[0xc869] = 0x00ef;    # small i with diaeresis
    $combined[0xc86f] = 0x00f6;    # small o with diaeresis
    $combined[0xc874] = 0x1e97;    # small t with diaeresis
    $combined[0xc875] = 0x00fc;    # small u with diaeresis
    $combined[0xc877] = 0x1e85;    # small w with diaeresis
    $combined[0xc878] = 0x1e8d;    # small x with diaeresis
    $combined[0xc879] = 0x00ff;    # small y with diaeresis
    # 4/9 umlaut
    $combined[0xc920] = 0x00a8;    # [diaeresis]
    $combined[0xc961] = 0x00e4;    # a with umlaut
    $combined[0xc965] = 0x00eb;    # e with umlaut
    $combined[0xc969] = 0x00ef;    # i with umlaut
    $combined[0xc96f] = 0x00f6;    # o with umlaut
    $combined[0xc975] = 0x00fc;    # u with umlaut
    # 4/10 circle above
    $combined[0xca41] = 0x00c5;    # capital a with ring above
    $combined[0xcaad] = 0x016e;    # capital u with ring above
    $combined[0xca61] = 0x00e5;    # small a with ring above
    $combined[0xca75] = 0x016f;    # small u with ring above
    $combined[0xca77] = 0x1e98;    # small w with ring above
    $combined[0xca79] = 0x1e99;    # small y with ring above
    # 4/11 high comma off centre
    # 4/12 inverted high comma centred
    # 4/13 double acute accent
    $combined[0xcd4f] = 0x0150;    # capital o with double acute
    $combined[0xcd55] = 0x0170;    # capital u with double acute
    $combined[0xcd6f] = 0x0151;    # small o with double acute
    $combined[0xcd75] = 0x0171;    # small u with double acute
    # 4/14 horn
    $combined[0xce54] = 0x01a0;    # latin capital letter o with horn
    $combined[0xce55] = 0x01af;    # latin capital letter u with horn
    $combined[0xce74] = 0x01a1;    # latin small letter o with horn
    $combined[0xce75] = 0x01b0;    # latin small letter u with horn
    # 4/15 caron (hacek
    $combined[0xcf41] = 0x01cd;    # capital a with caron
    $combined[0xcf43] = 0x010c;    # capital c with caron
    $combined[0xcf44] = 0x010e;    # capital d with caron
    $combined[0xcf45] = 0x011a;    # capital e with caron
    $combined[0xcf47] = 0x01e6;    # capital g with caron
    $combined[0xcf49] = 0x01cf;    # capital i with caron
    $combined[0xcf4b] = 0x01e8;    # capital k with caron
    $combined[0xcf4c] = 0x013d;    # capital l with caron
    $combined[0xcf4e] = 0x0147;    # capital n with caron
    $combined[0xcf4f] = 0x01d1;    # capital o with caron
    $combined[0xcf52] = 0x0158;    # capital r with caron
    $combined[0xcf53] = 0x0160;    # capital s with caron
    $combined[0xcf54] = 0x0164;    # capital t with caron
    $combined[0xcf55] = 0x01d3;    # capital u with caron
    $combined[0xcf5a] = 0x017d;    # capital z with caron
    $combined[0xcf61] = 0x01ce;    # small a with caron
    $combined[0xcf63] = 0x010d;    # small c with caron
    $combined[0xcf64] = 0x010f;    # small d with caron
    $combined[0xcf65] = 0x011b;    # small e with caron
    $combined[0xcf67] = 0x01e7;    # small g with caron
    $combined[0xcf69] = 0x01d0;    # small i with caron
    $combined[0xcf6a] = 0x01f0;    # small j with caron
    $combined[0xcf6b] = 0x01e9;    # small k with caron
    $combined[0xcf6c] = 0x013e;    # small l with caron
    $combined[0xcf6e] = 0x0148;    # small n with caron
    $combined[0xcf6f] = 0x01d2;    # small o with caron
    $combined[0xcf72] = 0x0159;    # small r with caron
    $combined[0xcf73] = 0x0161;    # small s with caron
    $combined[0xcf74] = 0x0165;    # small t with caron
    $combined[0xcf75] = 0x01d4;    # small u with caron
    $combined[0xcf7a] = 0x017e;    # small z with caron
    # 5/0 cedilla
    $combined[0xd020] = 0x00b8;    # cedilla
    $combined[0xd043] = 0x00c7;    # capital c with cedilla
    $combined[0xd044] = 0x1e10;    # capital d with cedilla
    $combined[0xd047] = 0x0122;    # capital g with cedilla
    $combined[0xd048] = 0x1e28;    # capital h with cedilla
    $combined[0xd04b] = 0x0136;    # capital k with cedilla
    $combined[0xd04c] = 0x013b;    # capital l with cedilla
    $combined[0xd04e] = 0x0145;    # capital n with cedilla
    $combined[0xd052] = 0x0156;    # capital r with cedilla
    $combined[0xd053] = 0x015e;    # capital s with cedilla
    $combined[0xd054] = 0x0162;    # capital t with cedilla
    $combined[0xd063] = 0x00e7;    # small c with cedilla
    $combined[0xd064] = 0x1e11;    # small d with cedilla
    $combined[0xd065] = 0x0119;    # small e with cedilla
    $combined[0xd067] = 0x0123;    # small g with cedilla
    $combined[0xd068] = 0x1e29;    # small h with cedilla
    $combined[0xd06b] = 0x0137;    # small k with cedilla
    $combined[0xd06c] = 0x013c;    # small l with cedilla
    $combined[0xd06e] = 0x0146;    # small n with cedilla
    $combined[0xd072] = 0x0157;    # small r with cedilla
    $combined[0xd073] = 0x015f;    # small s with cedilla
    $combined[0xd074] = 0x0163;    # small t with cedilla
    # 5/1 rude
    # 5/2 hook to left
    # 5/3 ogonek (hook to right
    $combined[0xd320] = 0x02db;    # ogonek
    $combined[0xd341] = 0x0104;    # capital a with ogonek
    $combined[0xd345] = 0x0118;    # capital e with ogonek
    $combined[0xd349] = 0x012e;    # capital i with ogonek
    $combined[0xd34f] = 0x01ea;    # capital o with ogonek
    $combined[0xd355] = 0x0172;    # capital u with ogonek
    $combined[0xd361] = 0x0105;    # small a with ogonek
    $combined[0xd365] = 0x0119;    # small e with ogonek
    $combined[0xd369] = 0x012f;    # small i with ogonek
    $combined[0xd36f] = 0x01eb;    # small o with ogonek
    $combined[0xd375] = 0x0173;    # small u with ogonek
    # 5/4 circle below
    $combined[0xd441] = 0x1e00;    # capital a with ring below
    $combined[0xd461] = 0x1e01;    # small a with ring below
    # 5/5 half circle below
    $combined[0xf948] = 0x1e2a;    # capital h with breve below
    $combined[0xf968] = 0x1e2b;    # small h with breve below
    # 5/6 dot below
    $combined[0xd641] = 0x1ea0;    # capital a with dot below
    $combined[0xd642] = 0x1e04;    # capital b with dot below
    $combined[0xd644] = 0x1e0c;    # capital d with dot below
    $combined[0xd645] = 0x1eb8;    # capital e with dot below
    $combined[0xd648] = 0x1e24;    # capital h with dot below
    $combined[0xd649] = 0x1eca;    # capital i with dot below
    $combined[0xd64b] = 0x1e32;    # capital k with dot below
    $combined[0xd64c] = 0x1e36;    # capital l with dot below
    $combined[0xd64d] = 0x1e42;    # capital m with dot below
    $combined[0xd64e] = 0x1e46;    # capital n with dot below
    $combined[0xd64f] = 0x1ecc;    # capital o with dot below
    $combined[0xd652] = 0x1e5a;    # capital r with dot below
    $combined[0xd653] = 0x1e62;    # capital s with dot below
    $combined[0xd654] = 0x1e6c;    # capital t with dot below
    $combined[0xd655] = 0x1ee4;    # capital u with dot below
    $combined[0xd656] = 0x1e7e;    # capital v with dot below
    $combined[0xd657] = 0x1e88;    # capital w with dot below
    $combined[0xd659] = 0x1ef4;    # capital y with dot below
    $combined[0xd65a] = 0x1e92;    # capital z with dot below
    $combined[0xd661] = 0x1ea1;    # small a with dot below
    $combined[0xd662] = 0x1e05;    # small b with dot below
    $combined[0xd664] = 0x1e0d;    # small d with dot below
    $combined[0xd665] = 0x1eb9;    # small e with dot below
    $combined[0xd668] = 0x1e25;    # small h with dot below
    $combined[0xd669] = 0x1ecb;    # small i with dot below
    $combined[0xd66b] = 0x1e33;    # small k with dot below
    $combined[0xd66c] = 0x1e37;    # small l with dot below
    $combined[0xd66d] = 0x1e43;    # small m with dot below
    $combined[0xd66e] = 0x1e47;    # small n with dot below
    $combined[0xd66f] = 0x1ecd;    # small o with dot below
    $combined[0xd672] = 0x1e5b;    # small r with dot below
    $combined[0xd673] = 0x1e63;    # small s with dot below
    $combined[0xd674] = 0x1e6d;    # small t with dot below
    $combined[0xd675] = 0x1ee5;    # small u with dot below
    $combined[0xd676] = 0x1e7f;    # small v with dot below
    $combined[0xd677] = 0x1e89;    # small w with dot below
    $combined[0xd679] = 0x1ef5;    # small y with dot below
    $combined[0xd67a] = 0x1e93;    # small z with dot below

    # 5/7 double dot below
    $combined[0xd755] = 0x1e72;    # capital u with diaeresis below
    $combined[0xd775] = 0x1e73;    # small u with diaeresis below

    $combined_chars = $c1 * 256 + $c2;
    if (array_key_exists($combined_chars, $combined))
      return html_entity_decode('&#'. $combined[$combined_chars] . ';', ENT_NOQUOTES, 'UTF-8');

    $single = [0xbd => 'ʹ',
               0xe1 => 'Æ',
               0xe2 => 'Đ',
               0xe6 => 'Ĳ',
               0xe8 => 'Ł',
               0xe9 => 'Ø',
               0xea => 'Œ',
               0xec => 'Þ',
               0xf1 => 'æ',
               0xf2 => 'đ',
               0xf3 => 'ð',
               0xf5 => 'ı',
               0xf6 => 'ĳ',
               0xf8 => 'ł',
               0xf9 => 'ø',
               0xfa => 'œ',
               0xfb => 'ß',
               0xfc => 'þ',
    ];

    return (array_key_exists($c1, $single) ? $single[$c1] : '?')
      . chr($c2);
  }
}